/*
 * Copyright (c) 2007 Tomi Aarnio
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

//----------------------------------------------------------------------
// Reports the number of individual M3G objects in a scene,
// and estimates their memory consumption.
//----------------------------------------------------------------------

package com.graphicsformasses.ch13;
import com.graphicsformasses.framework.DemoTask;
import com.graphicsformasses.tools.SceneAnalyzer;

import java.util.*;
import javax.microedition.m3g.*;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;
import javax.microedition.lcdui.Font;

public class MemUsageEstimate extends DemoTask
{
    private World myWorld;
    private boolean redraw;
    private int scrollLines;
    private int scrollTime;

    private long reportedSize;
    private int[] numInstances = new int[classNames.length];
    private int[] estimatedSize = new int[classNames.length];
    
    private static String[] classNames = {
        "Object3D              ",
        "  AnimationController ",
        "  AnimationTrack      ",
        "  Appearance          ",
        "  Background          ",
        "  CompositingMode     ",
        "  Fog                 ",
        "  Image2D             ",
        "  IndexBuffer         ",
        "  KeyframeSequence    ",
        "  Material            ",
        "  PolygonMode         ",
        "  Transformable       ",
        "    Node              ",
        "      Camera          ",
        "      Group           ",
        "        World         ",
        "      Light           ",
        "      Mesh            ",
        "        MorphingMesh  ",
        "        SkinnedMesh   ",
        "      Sprite3D        ",
        "    Texture2D         ",
        "  VertexArray         ",
        "  VertexBuffer        ",
    };

    protected String getName() {
        return "MemUsageEstimate";
    }

    protected void start() {
        try {
            long freeMem = Runtime.getRuntime().freeMemory();
            long totalMem = Runtime.getRuntime().totalMemory();
            myWorld = (World)Loader.load("/pond.m3g")[0];
            freeMem -= Runtime.getRuntime().freeMemory();
            totalMem = Runtime.getRuntime().totalMemory() - totalMem;
            reportedSize = freeMem + totalMem;
            countObjects(myWorld, classNames, numInstances, estimatedSize);
            redraw = true;
        } catch (Exception e) {
            System.out.println(e.getMessage());
        }
    }

    protected void countObjects(Object3D root,
                                String[] classNames,
                                int[] numInstances,
                                int[] estimatedSize) throws ClassNotFoundException {
        
        Class[] classes = new Class[classNames.length];
        Object3D[] objects = new Object3D[100];
        
        for (int i=0; i < classes.length; i++) {
            try {
                classes[i] = Class.forName("javax.microedition.m3g." + classNames[i].trim());
                numInstances[i] = SceneAnalyzer.findUniqueObjects(objects, myWorld, classes[i]);
                for (int k=0; k < numInstances[i]; k++) {
                    estimatedSize[i] += SceneAnalyzer.memSize(objects[k]);
                }
                for (int j=0; j < objects.length; j++) objects[j] = null;
            }
            catch (IndexOutOfBoundsException e) {
                objects = new Object3D[objects.length*2];
                i--;
            }
        }
    }

    protected boolean run(int time, int keyStates, int keysChanged, Graphics g) {

        boolean scrollUp   = (keyStates & GameCanvas.UP_PRESSED) != 0;
        boolean scrollDown = (keyStates & GameCanvas.DOWN_PRESSED) != 0;
        
        if (time - scrollTime > 100) {
            if (scrollUp) {
                if (scrollLines > 0) {
                    --scrollLines;
                    redraw = true;
                }
            }
            if (scrollDown) {
                if (scrollLines < classNames.length - 1) {
                    ++scrollLines;
                    redraw = true;
                }
            }
            if (scrollUp || scrollDown) {
                scrollTime = time;
            }
        }
        if (!(scrollUp || scrollDown)) {
            scrollTime = 0;
        }
        
        
        if (redraw) {
            g.setColor(0x00000000);
            g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
            g.setColor(0xFF44FF55);

            g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_SMALL));
            g.drawString("Class", 0, 0, Graphics.TOP|Graphics.LEFT);
            g.drawString("#", g.getClipWidth()*3/4, 0, Graphics.TOP|Graphics.RIGHT);
            g.drawString("kB", g.getClipWidth(), 0, Graphics.TOP|Graphics.RIGHT);
            int y = g.getFont().getHeight();
            
            g.setFont(Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_SMALL));
            int h = g.getFont().getHeight();
            y = h;

            for (int i=scrollLines; i < classNames.length; i++) {
                g.drawString(classNames[i], 0, y, Graphics.TOP|Graphics.LEFT);
                g.drawString(" " + numInstances[i], g.getClipWidth()*3/4, y, Graphics.TOP|Graphics.RIGHT);
                g.drawString(" " + (estimatedSize[i]+512)/1024, g.getClipWidth(), y, Graphics.TOP|Graphics.RIGHT);
                y += h;
            }

            g.drawString("Reported by freeMemory(): ", 0, y+h, Graphics.TOP|Graphics.LEFT);
            g.drawString(" " + (reportedSize+512)/1024, g.getClipWidth(), y+h, Graphics.TOP|Graphics.RIGHT);

            redraw = false;
        }
        return true;
    }

}
