/*
 * Copyright (c) 2007 Kimmo Roimela
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.ch14;
import com.graphicsformasses.framework.DemoTask;
import com.graphicsformasses.tools.ShapeFactory;

import javax.microedition.m3g.*;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.Font;

public class GlossTexture extends DemoTask
{
    private Graphics3D g3d = Graphics3D.getInstance();
    
    private Camera camera = new Camera();
    private World scene = new World();
    
    private Node earth;
    private Node sun;

    private Font smallFont = Font.getFont(Font.FACE_PROPORTIONAL,
                                          Font.STYLE_PLAIN,
                                          Font.SIZE_SMALL);
    
    private int screenWidth, screenHeight;
    
    protected String getName() {
        return "Gloss mapping";
    }
    
    protected void resize(int w, int h) {
        camera.setPerspective(60.0f, (float)w/(float)h, 100.0f, 30000.0f);
        screenWidth = w;
        screenHeight = h;
    }
        
    protected void start() {

        Image earthImage;
        
        try {
            earthImage = Image.createImage("/earth_clouds.png");
        }
        catch (Exception e) {
            throw new Error("Texture not found");
        }

        // First pass applies diffuse lighting modulated by the texture
        
        Appearance diffusePass = new Appearance();
        
        Image2D myTextureColor = new Image2D(Image2D.RGB, earthImage);
        Texture2D diffuseTexture = new Texture2D(myTextureColor);
        diffuseTexture.setFiltering(Texture2D.FILTER_NEAREST,
                                    Texture2D.FILTER_NEAREST);
        diffusePass.setTexture(0, diffuseTexture);
        
        Material diffuseMaterial = new Material();
        diffuseMaterial.setColor(Material.AMBIENT|Material.DIFFUSE,
                                 0xFFF0F2FF);
        diffuseMaterial.setColor(Material.EMISSIVE, 0x191630);
        diffusePass.setMaterial(diffuseMaterial);
        
        // Second pass adds specular lighting on top of the previous pass
        
        Appearance specularPass = new Appearance();
        
        Image2D myTextureGloss = new Image2D(Image2D.ALPHA, earthImage);
        Texture2D specularTexture = new Texture2D(myTextureGloss);
        specularTexture.setFiltering(Texture2D.FILTER_NEAREST,
                                     Texture2D.FILTER_NEAREST);
        specularTexture.setBlending(Texture2D.FUNC_REPLACE);
        specularPass.setTexture(0, specularTexture);
        
        Material specularMaterial = new Material();
        specularMaterial.setColor(Material.AMBIENT|Material.DIFFUSE, 0);
        specularMaterial.setColor(Material.SPECULAR, 0xE0CA80);
        specularMaterial.setShininess(80.f);
        specularPass.setMaterial(specularMaterial);
        
        CompositingMode additiveAlphaBlend = new CompositingMode();
        additiveAlphaBlend.setBlending(CompositingMode.ALPHA_ADD);
        specularPass.setCompositingMode(additiveAlphaBlend);

        // Build a Mesh object incorporating both passes
        
        Mesh earthSphere = ShapeFactory.makeSphere(12742.f, 3000);
        
        VertexBuffer vb = earthSphere.getVertexBuffer();
        vb.setDefaultColor(0xFFFF0000);
        
        IndexBuffer[] ib = new IndexBuffer[2];
        ib[0] = earthSphere.getIndexBuffer(0);
        ib[1] = ib[0];
        
        Appearance[] app = new Appearance[2];
        app[0] = diffusePass;
        app[1] = specularPass;
        
        earth = new Mesh(vb, ib, app);
        earth.setOrientation(-60.f, 1.f, 0.f, 0.f);
        scene.addChild(earth);

        sun = new Light();
        ((Light)sun).setColor(0xCBC1B2);
        ((Light)sun).setIntensity(2.0f);
        sun.setOrientation(120.f, 0.f, 1.f, 0.f);
        scene.addChild(sun);

        camera.translate(0.f, 1000.f, 20000.f);
        camera.setOrientation(-10.f, 1.f, 0.f, 0.f);
        scene.addChild(camera);
        scene.setActiveCamera(camera);
    }

    protected boolean run(int time,
                          int keyStates, int keysChanged,
                          Graphics g) {
        try {
            earth.postRotate(2.f, 0.f, 0.f, 1.f);
            sun.postRotate(-1.f, 0.f, 1.f, 0.f);
            
            g3d.bindTarget(g);
            g3d.render(scene);
        }
        finally {
            g3d.releaseTarget();
        }

        g.setFont(smallFont);
        g.setColor(0x555555);
        g.drawString("Image from NASA Visible Earth", 2, screenHeight-2, Graphics.BOTTOM|Graphics.LEFT);
        
        return true;
    }
}
