/*
 * Copyright (c) 2007 Kimmo Roimela
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

//----------------------------------------------------------------------
// Draw a simple skinned character
//----------------------------------------------------------------------

package com.graphicsformasses.ch16;
import com.graphicsformasses.framework.DemoTask;

import java.lang.Math;
import javax.microedition.lcdui.Graphics;
import javax.microedition.m3g.*;

public class Skinning extends DemoTask
{
    //------------------------------------------------------------------
    // Mesh data
    //------------------------------------------------------------------
    
    static private byte vertices[] = {
        // Head and neck
        -10, 127, 0,  10, 127, 0,  -25, 103, 0,  25, 103, 0,
        -10,  80, 0,  10,  80, 0,  -15,  70, 0,  15,  70, 0,
        // Arms and torso
        -120, 55, 0,  -120, 40, 0,  -80, 60, 0,  -80, 40, 0,  -40, 65, 0,
         -40, 40, 0,   120, 55, 0,  120, 40, 0,   80, 60, 0,   80, 40, 0,
          40, 65, 0,    40, 40, 0,  -20, 0, 0,    20, 0, 0,
        // Lower body and legs
        -30,  -20, 0,    0,  -30, 0,   30, -20, 0,  -30, -60, 0,
         -5,  -60, 0,  -30, -120, 0,  -10, -120, 0,   5, -60, 0,
         30, -60, 0,    10, -120, 0,   30, -120, 0
    };

    static private int tristrips[] = {
        1, 0, 3, 2, 5, 4, 7, 6,
        8, 9, 10, 11, 12, 13, 6, 20, 7, 21, 18, 19, 16, 17, 14, 15,
        20, 22, 21, 23, 24,
        23, 22, 26, 25, 28, 27,
        24, 23, 30, 29, 32, 31
    };

    static private int striplens[] = {
        8, 16, 5, 6, 6
    };
    
    //------------------------------------------------------------------
    // Instance variables
    //------------------------------------------------------------------    
    
    private Graphics3D g3d = Graphics3D.getInstance();
    private Background background = new Background();
    private Camera camera = new Camera();

    private SkinnedMesh stickMan;
    private Group torso, neck, head;
    private Group leftThigh, rightThigh, leftShin, rightShin;
    private Group leftUpperArm, rightUpperArm, leftForeArm, rightForeArm;
    
    protected String getName() {
        return "Skinning";
    }

    protected String getDescription() {
        return "";
    }

    //
    // Build the SkinnedMesh object
    //
    protected void start() {

        // Create the vertices and triangle strips for the mesh
        
        VertexArray pos = new VertexArray(vertices.length / 3, 3, 1);
        pos.set(0, vertices.length / 3, vertices);

        VertexBuffer vb = new VertexBuffer();
        vb.setPositions(pos, 1.f, null);
        vb.setDefaultColor(0x000000);
        
        IndexBuffer ib = new TriangleStripArray(tristrips, striplens);

        // Create the skinned mesh and set up its skeleton hierarchy
        
        stickMan = new SkinnedMesh(vb, ib, new Appearance(), new Group());
        Group pelvis = stickMan.getSkeleton();

        // Connect the torso, neck, and head
        
        torso = new Group();
        pelvis.addChild(torso);

        neck = new Group();
        torso.addChild(neck);
        neck.setTranslation(0.f, 60.f, 0.f);

        head = new Group();
        neck.addChild(head);
        head.setTranslation(0.f, 20.f, 0.f);

        // Connect the arms to the torso
        
        leftUpperArm = new Group();
        torso.addChild(leftUpperArm);
        leftUpperArm.setTranslation(30.f, 50.f, 0.f);
        leftUpperArm.setOrientation(-90.f, 0.f, 0.f, 1.f);

        leftForeArm = new Group();
        leftUpperArm.addChild(leftForeArm);
        leftForeArm.setTranslation(0.f, 50.f, 0.f);

        rightUpperArm = new Group();
        torso.addChild(rightUpperArm);
        rightUpperArm.setTranslation(-30.f, 50.f, 0.f);
        rightUpperArm.setOrientation(90.f, 0.f, 0.f, 1.f);

        rightForeArm = new Group();
        rightUpperArm.addChild(rightForeArm);
        rightForeArm.setTranslation(0.f, 50.f, 0.f);

        // Connect the legs to the 
        leftThigh = new Group();
        pelvis.addChild(leftThigh);
        leftThigh.setTranslation(20.f, -15.f, 0.f);
        leftThigh.setOrientation(180.f, 0.f, 0.f, 1.f);

        leftShin = new Group();
        leftThigh.addChild(leftShin);
        leftShin.setTranslation(0.f, 45.f, 0.f);                                   
        rightThigh = new Group();
        pelvis.addChild(rightThigh);
        rightThigh.setTranslation(-20.f, -15.f, 0.f);
        rightThigh.setOrientation(180.f, 0.f, 0.f, 1.f);

        rightShin = new Group();
        rightThigh.addChild(rightShin);
        rightShin.setTranslation(0.f, 45.f, 0.f);

        // Set bone influences on vertices
        
        stickMan.addTransform(torso, 100, 6, 2);
        stickMan.addTransform(torso, 100, 12, 2);        
        stickMan.addTransform(torso, 100, 18, 2);
        stickMan.addTransform(torso, 100, 20, 2);

        stickMan.addTransform(neck, 30, 4, 4);
        stickMan.addTransform(head, 100, 0, 6);

        stickMan.addTransform(rightForeArm, 100, 8, 4);
        stickMan.addTransform(rightUpperArm, 100, 10, 4);

        stickMan.addTransform(leftForeArm, 100, 14, 4);
        stickMan.addTransform(leftUpperArm, 100, 16, 4);

        stickMan.addTransform(leftThigh, 100, 23, 2);
        stickMan.addTransform(leftThigh, 100, 29, 2);
        stickMan.addTransform(leftShin, 100, 29, 4);
        
        stickMan.addTransform(rightThigh, 100, 22, 2);
        stickMan.addTransform(rightThigh, 100, 25, 2);
        stickMan.addTransform(rightShin, 100, 25, 4);

        // Rotate the legs to a default position -- note that when
        // called after addTransform, this has no effect on the bone
        // influences

        leftThigh.postRotate(10.f, 0.f, 0.f, 1.f);
        rightThigh.postRotate(-10.f, 0.f, 0.f, 1.f);

        leftUpperArm.postRotate(-80.f, 0.f, 0.f, 1.f);
        rightUpperArm.postRotate(-60.f, 0.f, 0.f, 1.f);
        rightForeArm.postRotate(-30.f, 0.f, 0.f, 1.f);

        // Initialize the rest of the objects
        
        g3d.setCamera(camera, null);
        background.setColor(0xFFCC22);
    }

    //
    // Reset camera and viewport
    //
    protected void resize(int width, int height) {
        camera.setParallel(400.f, (float) width / (float) height, -10.f, 10.f);
    }

    //
    // Animate and render
    //
    protected boolean run(int time, int keyStates, int keysChanged, Graphics g) {

        try {
            g3d.bindTarget(g);
            g3d.clear(background);
            g3d.render(stickMan, null);
        }
        finally {
            g3d.releaseTarget();
        }
    
        return true;
    }

    protected void stop() {
    }
}
