/*
 * Copyright (c) 2007 Kimmo Roimela
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.framework;

import java.lang.System;
import java.util.Timer;
import java.util.TimerTask;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.GameCanvas;
    
/**
 * A Canvas class for displaying and switching between a collection of
 * demos implemented as DemoTask instances.  A DemoCanvas builds a
 * menu for presenting a collection of DemoTasks from a given package
 * and controls their execution.
 *
 * @see DemoTask
 */
class DemoCanvas extends GameCanvas implements CommandListener, Runnable
{
    class DemoTimer extends TimerTask
    {
        private DemoCanvas canvas;
        
        DemoTimer(DemoCanvas canvas) {
            this.canvas = canvas;
        }
                
        public void run() {
            canvas.run();
        }
    }
    
    private DemoMIDlet midlet;
    private String demoPackage;
    private DemoTask task;
    private Graphics g;
    private Display display;

    private Timer timer;
    private long startTime;

    private int lastKeyStates;
    private boolean running;

    //------------------------------------------------------------------
    // Construction
    //------------------------------------------------------------------
    
    /**
     * Returns a new DemoCanvas instance
     *
     * @param parent       the host MIDlet displaying this canvas
     * @param demoPackage  Java package name string for the demos
     * @param demos        array of demo class name strings
     */
    protected DemoCanvas(DemoMIDlet parent, String demoPackage, String[] demos) {
        super(true);
        midlet = parent;
        this.demoPackage = demoPackage;
        display = Display.getDisplay(parent);
        g = getGraphics();

        // Add our demos as menu commands

        for (int i = 0; i < demos.length; ++i) {
            addCommand(new Command(demos[i], Command.SCREEN, 2));
        }
        addCommand(new Command("Exit", Command.EXIT, 1));
        setCommandListener(this);
    }

    //------------------------------------------------------------------
    // Package functionality
    //------------------------------------------------------------------
    
    /**
     * (Re)starts the current demo.
     */
    void start() {
        if (task != null) {
            setFullScreenMode(true);
            task.start();
            task.resize(getWidth(), getHeight());
            g.setClip(0, 0, getWidth(), getHeight());
            running = true;

            startTime = System.currentTimeMillis();
            //run()
            timer = new Timer();
            timer.schedule(new DemoTimer(this), 300, 50);
        }
    }

    /**
     * Starts a new demo.
     *
     * @param demo  the DemoTask to start
     */
    void start(DemoTask demo) {
        stop();
        task = demo;
        start();
    }

    /**
     * Starts a new demo based on class name.
     *
     * @param demoClassName class name of the DemoTask to start,
     * omitting the package name
     */
    void start(String demoClassName) {
        try {
            Class cls = Class.forName(demoPackage + "." + demoClassName);
            start((DemoTask) cls.newInstance());
        }
        catch (Exception e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
    }
    
    /**
     * Executes the current DemoTask for one cycle.
     */
    public void run() {
        if (running) {
            int demoTime = (int)(System.currentTimeMillis() - startTime);
            int keyStates = getKeyStates();
            boolean stayAlive = false;
            try {
                stayAlive = task.run(demoTime, keyStates, keyStates ^ lastKeyStates, g);
            }
            catch (Exception e) {
                System.out.println(e.getMessage());
                e.printStackTrace();
                task.stop();
            }
            lastKeyStates = keyStates;
    
            flushGraphics();
            
            if (stayAlive) {
                //display.callSerially(this);
            }
            else {
                stop();
            }
        }
    }

    /**
     * Stops the current demo.
     */
    void stop() {
        if (task != null) {
            if (timer != null) {
                timer.cancel();
            }
            timer = null;
            task.stop();
        }
        setFullScreenMode(false);
        running = false;
    }

    //------------------------------------------------------------------
    // Event handlers
    //------------------------------------------------------------------
    
    //
    // Handle the menu commands
    //
    public void commandAction(Command c, Displayable d) {
        if (c.getLabel() == "Exit") {
            midlet.destroyApp(true);
            return;
        }
        start(c.getLabel());
    }

    //
    // Handle key presses
    //
    protected void keyPressed(int keyCode) {
    }
    protected void keyRepeated(int keyCode) {
    }
    protected void keyReleased(int keyCode) {
    }
    
    //
    // Handle repaints
    //
    public void paint(Graphics g) {
        if (!running) {
            flushGraphics();
        }
    }

    //
    // Signal size change to the demo
    //
    protected void sizeChanged(int w, int h) {
        g.setClip(0, 0, getWidth(), getHeight());
        if (task != null) {
            task.resize(w, h);
        }
    }
}
