/*
 * Copyright (c) 2007 Tomi Aarnio
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.tools;

import javax.microedition.m3g.*;

/*-----------------------------------------------------*
 * A class providing read access to KeyframeSequences. *
 * Uses exceptions to infer the otherwise hidden data. *
 *-----------------------------------------------------*/

public class KeyframeSequenceReader
{
    int componentCount;
    int keyframeCount;
    int interpolation;
    int validRangeFirst;
    int validRangeLast;

    int kftimes[];

    KeyframeSequence dupe;
    MorphingMesh dummyTarget;

    public KeyframeSequenceReader(KeyframeSequence original)
    {
        dupe = (KeyframeSequence)original.duplicate();
        KeyframeSequence temp = (KeyframeSequence)original.duplicate();

        // Extract the number of keyframes
        //
        keyframeCount = 1;
        while (true) {
            try {
                temp.setValidRange(0, keyframeCount);
                keyframeCount++;
            } catch (IndexOutOfBoundsException e) {
                break;
            }
        }
        kftimes = new int[keyframeCount];

    	// Extract the number of components per keyframe
        //
        componentCount = 1;
        while (true) {
            float[] components = new float[componentCount];
            try {
                temp.setKeyframe(0, 0, components);
                break;
            } catch (IllegalArgumentException e) {
                componentCount++;
            }
        }

        // Set up a dummy MorphingMesh to receive the keyframe values
        //
        VertexBuffer vb = new VertexBuffer();
        IndexBuffer ib = new TriangleStripArray(0, new int[] {3});
        VertexBuffer targets[] = new VertexBuffer[componentCount];
        IndexBuffer submeshes[] = new IndexBuffer[componentCount];
        for (int i=0; i < componentCount; i++) {
            targets[i] = vb;
            submeshes[i] = ib;
        }
        dummyTarget = new MorphingMesh(vb, targets, submeshes, null);
        AnimationTrack dummyTrack = 
            new AnimationTrack(dupe, AnimationTrack.MORPH_WEIGHTS);
        dummyTrack.setController(new AnimationController());
        dummyTarget.addAnimationTrack(dummyTrack);

        // Fetch the keyframe time stamps, store them into an array.
        //
        // LIMITATION: If the time stamp of a keyframe is zero, it
        // will be reported as being one.  We cannot distinguish a
        // time stamp of zero from one, because setDuration does not
        // allow the duration to be set to zero.
        //
        int kfindex=0;
        int duration=1;
        while (kfindex < getKeyframeCount()) {
            dupe.setValidRange(kfindex, kfindex);
            dupe.setDuration(duration);
            try {
                dummyTarget.animate(0);
                kftimes[kfindex++] = duration;
            } catch (IllegalStateException e) {
                duration++;
            }
        }

        // Work around the limitation described above. After this,
        // we know the correct time stamp of each and every keyframe.
        //
        for (kfindex=0; kftimes[kfindex]==1; kfindex++);
        if (kfindex > 0) {
            float[] tempkf = new float[this.getComponentCount()];
            this.getKeyframe(kfindex, tempkf);
            dupe.setDuration(1);
            dupe.setValidRange(0, kfindex);
            dupe.setKeyframe(kfindex, 0, tempkf);
            try {
                dummyTarget.animate(0);
                for (int i=0; i < kfindex; i++) {
                    kftimes[i] = 0;
                }
            } catch (IllegalStateException e) {
            }
            dupe.setKeyframe(kfindex, kftimes[kfindex], tempkf);
        }

        // Detect the interpolation mode.
        //
        // Only STEP and LINEAR have been reliably tested.
        //
        float[] backup0 = new float[this.getComponentCount()];
        float[] backup1 = new float[this.getComponentCount()];
        dupe.getKeyframe(0, backup0);
        dupe.getKeyframe(1, backup1);

        float[] tempkf = new float[this.getComponentCount()];
        tempkf[0] = 0.0f;
        dupe.setKeyframe(0, 0, tempkf);
        tempkf[0] = 10.0f;
        dupe.setKeyframe(1, 10, tempkf);
        dupe.setRepeatMode(KeyframeSequence.CONSTANT);
        dupe.setValidRange(0, 1);
        dupe.setDuration(10);
        dummyTarget.animate(1);
        dummyTarget.getWeights(tempkf);
        System.out.println("Detecting interpolation mode, tempkf[0] = " +tempkf[0]);
        if (tempkf[0] < 0.001f && tempkf[0] > -0.001f)
            interpolation = KeyframeSequence.STEP;
        else if (tempkf[0] < 1.001f && tempkf[0] > 0.999f)
            interpolation = KeyframeSequence.LINEAR;
        else if (tempkf[0] < 0.2f && tempkf[0] > 0.1f)
            interpolation = KeyframeSequence.SLERP;
        else if (tempkf[0] < 0.3f && tempkf[0] > 0.2f)
            interpolation = KeyframeSequence.SPLINE;
        else
            interpolation = KeyframeSequence.SQUAD;

        dupe.setKeyframe(0, kftimes[0], backup0);
        dupe.setKeyframe(1, kftimes[1], backup1);

        // Finally restore the original duration and valid range.
        //
        dupe.setDuration(original.getDuration());
        dupe.setRepeatMode(original.getRepeatMode());
        dupe.setValidRange(0, this.getKeyframeCount()-1);
    }

    public int getComponentCount() {
        return componentCount;
    }

    public int getKeyframeCount() {
        return keyframeCount;
    }

    public int getInterpolationType() {
        return interpolation;
    }

    public int getKeyframe(int index, float[] value)
    {
        if (index < 0 || index >= getKeyframeCount())
            throw new IndexOutOfBoundsException();

        if (value != null && value.length < getComponentCount())
            throw new IllegalArgumentException();

        dupe.setValidRange(index, index);
        dummyTarget.animate(0);
        dummyTarget.getWeights(value);
        return kftimes[index];
    }
}
