/*
 * Copyright (c) 2007 Kimmo Roimela
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.tools;

import javax.microedition.m3g.AnimationController;
import javax.microedition.m3g.KeyframeSequence;

/**
 * Abstract base class for streaming animation keyframes.
 */
abstract public class KeyframeStream
{
    private AnimationController myController;
    private KeyframeSequence myBuffer;

    private int interpolation;

    private float[] nextKeyframe;
    private int secondLastKeyframeTime;

    /**
     * Class constructor.  Associates this keyframe stream with an
     * animation controller and a keyframe sequence.
     *
     * @param ctrl  AnimationController to use for timing information
     * @param seq   KeyframeSequence to stream keyframes into; must have
     *              at least four keyframes
     */
    protected KeyframeStream(AnimationController ctrl, KeyframeSequence seq) {

        if (seq.getKeyframeCount() < 4) {
            throw new IllegalArgumentException("Too few keyframes for buffer");
        }
        
        myController = ctrl;
        myBuffer = seq;
        myBuffer.setDuration(1); // a dirty trick to validate the sequence...

        interpolation = myBuffer.getInterpolationType();

        nextKeyframe = new float[seq.getComponentCount()];
    }

    /**
     * Updates the contents of the associated keyframe sequence.  New
     * keyframes are read into the buffer as required for
     * interpolating up until the given time.
     *
     * @param worldTime  the world time to animate to
     */
    public void update(int worldTime) {

        float seqTime = myController.getPosition(worldTime);
        
        while (seqTime >= myBuffer.getDuration()) {
            bufferNextKeyframe();
        }

        if (interpolation == KeyframeSequence.SPLINE || interpolation == KeyframeSequence.SQUAD) {
            if (seqTime >= secondLastKeyframeTime) {
                bufferNextKeyframe();
            }
        }
    }

    /**
     * Called to fetch the next keyframe in the stream.
     *
     * @param kf  a float vector to fill with new keyframe data
     * @return the time stamp (sequence time) of the new keyframe
     */
    abstract protected int getNextKeyframe(float[] kf);

    /**
     * Appends a new keyframe to the sequence and updates all the
     * necessary state variables.
     */
    private void bufferNextKeyframe() {
        int firstValid = myBuffer.getValidRangeFirst();
        int lastValid = firstValid;
        
        int nextKeyframeTime = getNextKeyframe(nextKeyframe);
        firstValid = (firstValid + 1) % myBuffer.getKeyframeCount();
        myBuffer.setValidRange(firstValid, lastValid);
        
        myBuffer.setKeyframe(lastValid, nextKeyframeTime, nextKeyframe);
        
        secondLastKeyframeTime = myBuffer.getDuration();
        myBuffer.setDuration(nextKeyframeTime);
    }
}
