/*
 * Copyright (c) 2007 Tomi Aarnio
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.tools;

import javax.microedition.m3g.*;

public abstract class SceneAnalyzer
{
    /**
     * Finds all Object3D instances that are reachable from the given
     * root object.  The returned objects are guaranteed to be unique,
     * but they are not in any particular order.
     *
     * This method is not very efficient: it takes O(N^2) time and
     * O(N) memory, where N is the number of objects traversed.  On
     * the other hand, finding objects in the scene graph is typically
     * a one-off process.
     * 
     * @param found an Object3D array to store the found objects;
     *        all elements must be initialized to null
     * @param root the Object3D to start searching from
     * @param type the Class whose instances to find
     *
     * @return the number of matching Object3Ds found
     *
     * @throws IndexOutOfBoundsException if the <tt>found</tt> array
     *         is not large enough to accommodate the matching Object3Ds
     */
    public static int findUniqueObjects(Object3D[] found,
                                        Object3D root,
                                        Class type)
    {
        int i, numUnique = 0;
    
        // Retrieve the scene graph objects that are directly referenced
        // by 'root' and process them recursively.  Note that we create
        // a new 'references' array at each step of the recursion; this
        // is not recommended as a general practice, but in this case we
        // are favoring brevity and clarity over efficiency.
        //
        Object3D[] references = new Object3D[root.getReferences(null)];
        root.getReferences(references);
        for (i=0; i < references.length; i++) {
            numUnique += findUniqueObjects(found, references[i], type);
        }

        // Check whether 'root' is an instance of 'type', and if so,
        // insert it at the end of the 'found' array, provided that it
        // has not been inserted before.  If there are no empty slots
        // at the end of the array, an IndexOutOfBoundsException will
        // be thrown.
        //
        if (type.isAssignableFrom(root.getClass())) {
            for (i=0; found[i] != root && found[i] != null;  i++);
            if (found[i] == null) {
                found[i] = root;
                numUnique++;
            }
        }
        return numUnique;
    }
    
    /**
     * Estimates the runtime memory consumption of the given Object3D.
     * 
     * The estimate is based solely on the raw data contained within
     * the object, and does not take the underlying M3G engine or the
     * VM into account.  For example, an RGB Image2D is estimated to
     * take 16 bits per pixel (bpp), although the actual memory usage
     * may be 8, 16, 24, or 32 bpp, depending on the device and
     * whether the image is input in palletized format.
     * 
     * @param object the Object3D whose memory consumption to estimate
     *
     * @return the estimated memory consumption in bytes
     */
    public static int memSize(Object3D object)
    {
        int size = 64;                                // assume 64 bytes for atomic data fields
        size += object.getAnimationTrackCount() * 4;  // 4 bytes per AnimationTrack reference

        if (object instanceof Image2D) {
            Image2D image = (Image2D) object;
            int bpp = 0;
            switch (image.getFormat()) {
            case Image2D.ALPHA:
                bpp = 1; break;
            case Image2D.LUMINANCE:
                bpp = 2; break;
            case Image2D.LUMINANCE_ALPHA:
                bpp = 2; break;
            case Image2D.RGB:
                bpp = 2; break;
            case Image2D.RGBA:
                bpp = 4; break;
            default:
            }
            size += image.getWidth() * image.getHeight() * bpp;
        }

        if (object instanceof IndexBuffer) {
            IndexBuffer indices = (IndexBuffer) object;
            size += indices.getIndexCount()*2;
        }

        if (object instanceof KeyframeSequence) {
            KeyframeSequence keyframes = (KeyframeSequence) object;
            int numKeyframes = keyframes.getKeyframeCount();
            int numComponents = keyframes.getKeyframeCount() * keyframes.getComponentCount();
            int timeStampBytes = keyframes.getKeyframeCount() * 4;
            int keyframeBytes;
            switch (keyframes.getInterpolationType()) {
            case KeyframeSequence.SPLINE:
                keyframeBytes = numComponents * 4;     // keyframe data [FLOAT]
                keyframeBytes += numComponents * 2*4;  // assume precomputed in/out tangents [FLOAT]
                break;
            case KeyframeSequence.SLERP:
                keyframeBytes = numComponents * 4;     // keyframe data [FLOAT]
                keyframeBytes += numKeyframes * 2*4;   // assume precomputed angle and sin(angle) [FLOAT]
                break;
            case KeyframeSequence.SQUAD:
                keyframeBytes = numComponents * 4;     // keyframe data [FLOAT]
                keyframeBytes += numKeyframes * 2*4*4; // two precomputed quaternions [FLOAT]
                break;
            default:
                keyframeBytes = numComponents * 4;     // keyframe data [FLOAT]
                break;
            }
            size += keyframeBytes + timeStampBytes;
        }

        if (object instanceof VertexArray) {
            VertexArray vertices = (VertexArray) object;
            size += vertices.getVertexCount() *
                    vertices.getComponentCount() *
                    vertices.getComponentType();
        }

        if (object instanceof Transformable) {
            size += (3 + 4 + 3*4) * 4;   // assume T R S M are stored as floats
            size += 3*4*4;               // assume one cached 3x4 float matrix
        }

        if (object instanceof Group) {
            Group group = (Group) object;
            size += group.getChildCount()*4;
        }

        if (object instanceof Camera) {
            size += (4 + 4*4)*4;        // 4 parameters + one 4x4 matrix
        }

        if (object instanceof Mesh) {
            Mesh mesh = (Mesh) object;
            size += mesh.getSubmeshCount()*2*4; // IB ref + Appearance ref
        }

        if (object instanceof MorphingMesh) {
            MorphingMesh mesh = (MorphingMesh) object;
            size += mesh.getMorphTargetCount()*2*4; // VB ref + weight, 4 bytes each
        }
        
        if (object instanceof SkinnedMesh) {
            SkinnedMesh mesh = (SkinnedMesh) object;
            size += mesh.getVertexBuffer().getVertexCount()*(4+4*2);
            // 4 bone indices (BYTE) + 4 weights (SHORT)
        }
        
        return size;
    }
    
}
