/*
 * Copyright (c) 2007 Kimmo Roimela
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.graphicsformasses.tools;

import java.lang.Math;

import javax.microedition.m3g.*;

/**
 * Factory class for creating meshes in simple shapes.  All created
 * shapes have normals and texture coordinates in addition to vertex
 * positions.
 */
public class ShapeFactory extends Object
{    
    static private final float[] defaultTexCoordBias = { 0.5f, 0.5f };

    private ShapeFactory() {}

    /**
     * Creates a cylinder Mesh.  The Mesh has three submeshes: 0 is
     * the main shell, and 1 and 2 are the end caps.
     * 
     * @param diameter      floating-point diameter of the cylinder
     * @param length        floating-point length of the cylinder
     * @param vertexBudget  integer number of vertices to (roughly) use
     *
     * @return the new Mesh object 
     */
    static public Mesh makeCylinder(float diameter, float length, int vertexBudget) {
        float radius = diameter / 2.f;
        int div = vertexBudget / 4;

        int vertexCount = div*4 + 2;
        
        VertexArray pos, normal = null, texCoord = null;
        pos = new VertexArray(vertexCount, 3, 2);
        normal = new VertexArray(vertexCount, 3, 1);
        texCoord = new VertexArray(vertexCount, 2, 1);

        float scale = 32767.f / Math.max(radius, length);

        short[] v = new short[4];
        byte[] n = new byte[3];
        byte[] tc = new byte[2];
        
        int vi = 0;
        
        // Make the main cylinder
        
        for (int i = 0; i <= div; ++i) {
            v[0] = (short)(scale * radius * Math.sin(2*Math.PI*i / div) + 0.5);
            v[1] = (short)(scale * radius * Math.cos(2*Math.PI*i / div) + 0.5);

            n[0] = (byte)(v[0] / 256.f);
            n[1] = (byte)(v[1] / 256.f);
            n[2] = 0;

            tc[0] = (byte)(255.f * (float)i / (float)div + 0.5f);
            
            v[2] = (short)(-scale * length / 2 + 0.5);
            tc[1] = (byte) -127;
            pos.set(vi, 1, v);
            normal.set(vi, 1, n);
            texCoord.set(vi, 1, tc);
            ++vi;
            
            v[2] = (short)(scale * length / 2 + 0.5);
            tc[1] = (byte) 127;
            pos.set(vi, 1, v);
            normal.set(vi, 1, n);
            texCoord.set(vi, 1, tc);
            ++vi;
        }
        
        IndexBuffer[] strips = new TriangleStripArray[3];
        int[] stripLens = new int[1];
        
        stripLens[0] = vi;
        strips[0] = new TriangleStripArray(0, stripLens);

        // Make the end caps

        n[0] = n[1] = (byte) 0;
        n[2] = (byte) -127;

        int bottomOffset = vi;
        
        for (int i = 0; i < div; ++i, ++vi) {
            v[0] = (short)(scale * radius * Math.sin(2*Math.PI*i / div) + 0.5);
            v[1] = (short)(scale * radius * Math.cos(2*Math.PI*i / div) + 0.5);
            v[2] = (short)(-scale * length / 2 + 0.5);
            
            tc[0] = (byte)(127.f * Math.sin(2*Math.PI*i / div) + 0.5f);
            tc[1] = (byte)(127.f * Math.cos(2*Math.PI*i / div) + 0.5f);
            
            pos.set(vi, 1, v);
            normal.set(vi, 1, n);
            texCoord.set(vi, 1, tc);
        }

        int topOffset = vi;
        
        n[2] = (byte) 127;
        
        for (int i = 0; i < div; ++i, ++vi) {
            v[0] = (short)(scale * radius * Math.sin(2*Math.PI*i / div) + 0.5);
            v[1] = (short)(scale * radius * Math.cos(2*Math.PI*i / div) + 0.5);
            v[2] = (short)(scale * length / 2 + 0.5);
            
            tc[0] = (byte)(127.f * Math.sin(2*Math.PI*i / div) + 0.5f);
            tc[1] = (byte)(127.f * Math.cos(2*Math.PI*i / div) + 0.5f);
            
            pos.set(vi, 1, v);
            normal.set(vi, 1, n);
            texCoord.set(vi, 1, tc);
        }

        int[] indicesTop    = new int[2*div];
        int[] indicesBottom = new int[2*div];
        
        for (int i = 0; i < div; ++i) {
            indicesTop[i*2]   = topOffset + i;
            indicesTop[i*2+1] = topOffset + div-1 - i;

            indicesBottom[i*2]   = bottomOffset + div-1 - i;
            indicesBottom[i*2+1] = bottomOffset + i;
        }
        
        stripLens[0] = 2*div;
        
        strips[1] = new TriangleStripArray(indicesTop, stripLens);
        strips[2] = new TriangleStripArray(indicesBottom, stripLens);
        
        // Create the VertexBuffer, Appearance and Mesh objects
        
        VertexBuffer vb = new VertexBuffer();
        vb.setPositions(pos, 1.0f / scale, null);
        vb.setNormals(normal);
        vb.setTexCoords(0, texCoord, 1.0f / 255.f, defaultTexCoordBias);

        Appearance[] apps = new Appearance[3];
        apps[0] = new Appearance();
        apps[1] = apps[2] = apps[0];
        
        return new Mesh(vb, strips, apps);
    }

    /**
     * Creates a sphere Mesh.
     *
     * @param diameter      floating-point diameter of the sphere
     * @param vertexBudget  integer number of vertices to (roughly) use
     * 
     * @return the new Mesh object
     */
    static public Mesh makeSphere(float diameter, int vertexBudget) {
        
        int rings = Math.max((int) Math.sqrt(vertexBudget / 2), 2);
        int div = Math.max((int) vertexBudget / (rings + 1), 2);

        int vertexCount = (rings+1)*(div+1);
        
        VertexArray pos, normal = null, texCoord = null;
        pos = new VertexArray(vertexCount, 3, 2);
        normal = new VertexArray(vertexCount, 3, 1);
        texCoord = new VertexArray(vertexCount, 2, 2);

        // Generate the sphere vertices
        
        short[] v  = new short[3];
        byte[]  n  = new byte[3];
        short[] tc = new short[2];
        
        for (int i = 0; i <= rings; ++i) {
            float z = (float) Math.cos(i * Math.PI / rings);
            float r = (float) Math.sin(i * Math.PI / rings);
            for (int j = 0; j <= div; ++j) {

                int vi = i*(div+1)+j;
                
                float x = (float)(-Math.sin(j * 2 * Math.PI / div) * r);
                float y = (float)( Math.cos(j * 2 * Math.PI / div) * r);

                v[0] = (short)(x * 32767.f + 0.5f);
                v[1] = (short)(y * 32767.f + 0.5f);
                v[2] = (short)(z * 32767.f + 0.5f);
                pos.set(vi, 1, v);

                n[0] = (byte)(x * 127.f + 0.5f);
                n[1] = (byte)(y * 127.f + 0.5f);
                n[2] = (byte)(z * 127.f + 0.5f);
                normal.set(vi, 1, n);

                tc[0] = (short)(32767.f * (float)j / (float)div + 0.5f);
                tc[1] = (short)(32767.f * (float)i / (float)rings + 0.5f);
                texCoord.set(vi, 1, tc);
            }
        }
        
        // Create the VertexBuffer, Appearance and Mesh objects

        int[] indices = new int[(rings*2)*(div+1)];
        int[] stripLen = new int[rings];
        int ii = 0;
        for (int i = 0; i < rings; ++i) {
            for (int j = 0; j <= div; ++j) {
                indices[ii++] = i*(div+1) + j;
                indices[ii++] = (i+1)*(div+1) + j;
            }
            stripLen[i] = 2*(div+1);
        }

        IndexBuffer strips = new TriangleStripArray(indices, stripLen);

        VertexBuffer vb = new VertexBuffer();
        vb.setPositions(pos, diameter / 65534.f, null);
        vb.setNormals(normal);
        vb.setTexCoords(0, texCoord, 1.0f / 32767.f, null);
        
        return new Mesh(vb, strips, new Appearance());
    }
    
}
